package scpt

import (
	"fmt"
	"github.com/alecthomas/participle/lexer"
)

// AST stores the root of the Abstract Syntax Tree for scpt
type AST struct {
	Pos      lexer.Position
	Commands []*Command `@@*`
}

// Execute traverses the AST and executes any commands, it returns an error
// containing the position at which the error was encountered and the error
// itself
func (ast *AST) Execute() error {
	for _, cmd := range ast.Commands {
		if cmd.Vars != nil {
			for _, Var := range cmd.Vars {
				val, err := ParseValue(Var.Value)
				if err != nil {
					return fmt.Errorf("%s: %s", Var.Value.Pos, err)
				}
				if IsFuncCall(val) {
					Call := val.(*FuncCall)
					Vars[Var.Key], err = CallFunction(Call)
					if err != nil {
						return fmt.Errorf("%s: %s", Var.Value.Pos, err)
					}
				} else {
					Vars[Var.Key] = val
				}
			}
		} else if cmd.Calls != nil {
			for _, Call := range cmd.Calls {
				_, err := CallFunction(Call)
				if err != nil {
					return fmt.Errorf("%s: %s", Call.Pos, err)
				}
			}
		}
	}
	return nil
}

// Command stores any commands encountered while parsing a script
type Command struct {
	Pos   lexer.Position
	Vars  []*Var      `( @@`
	Calls []*FuncCall `| @@ )`
}

// FuncCall stores any function calls encountered while parsing a script
type FuncCall struct {
	Pos  lexer.Position
	Name string `@Ident @("-" Ident)*`
	Args []*Arg `@@*`
}

// Arg stores arguments for function calls
type Arg struct {
	Pos   lexer.Position
	Key   string `("with" @Ident)?`
	Value *Value `@@`
}

// Var stores any variables encountered while parsing a script
type Var struct {
	Pos   lexer.Position
	Key   string `"set" @Ident "to"`
	Value *Value `@@`
}

// Value stores any literal values encountered while parsing a script
type Value struct {
	Pos     lexer.Position
	String  *string     `  @String`
	Float   *float64    `| @Float`
	Integer *int64      `| @Int`
	Bool    *Bool       `| @("true" | "false")`
	SubCmd  *FuncCall   `| "(" @@ ")"`
	VarVal  *string     `| "$" @Ident`
	Expr    *Expression `| "{" @@ "}"`
}

// Bool stores boolean values encountered while parsing a script.
// It is required for the Capture method
type Bool bool

// Capture parses a boolean literal encountered in the script into
// a Go boolean value
func (b *Bool) Capture(values []string) error {
	*b = values[0] == "true"
	return nil
}

// Expression stores any expressions encountered while parsing a
// script for later evaluation
type Expression struct {
	Pos   lexer.Position
	Left  *Value `@@`
	Op    string `@( ">" | ">" "=" | "<" | "<" "=" | "!" "=" | "=" "=" | "+" | "-" | "*" | "/" | "^" | "%")`
	Right *Value `@@`
}
